version "2.4"

class FlashlightPlusHolder : Inventory {
	FlashlightPlusLight light1;
	FlashlightPlusLight light2;
	bool on;
	
	void Enable(){
		PlayerPawn pp=PlayerPawn(owner);
		if(pp){
			if(light1){
				light1.destroy();
			}
			light1=FlashlightPlusLight(owner.Spawn("FlashlightPlusLight")).Init(pp,false);
			if(CVar.GetCVar("cl_flashlight_plus_second_beam",owner.player).GetBool()){
				if(light2){
					light2.destroy();
				}
				light2=FlashlightPlusLight(owner.Spawn("FlashlightPlusLight")).Init(pp,true);
			}
		}
		on=true;
	}
	
	void Disable(){
		if(light1){
			light1.destroy();
			light1=null;
		}
		if(light2){
			light2.destroy();
			light2=null;
		}
		on=false;
	}
	
	FlashlightPlusHolder Init() {
		light1=null;
		light2=null;
		on=false;
		return self;
	}
	
	void ToggleFlashlight(){
		if(on){
			Disable();
		}else{
			Enable();
		}
	}
	
	void FixState(){
		if(!owner){
			destroy();
		}else{
			if(on){
				Enable();
			}else{
				Disable();
			}
		}
	}
}

class FlashLightPlusHandler : StaticEventHandler {
	
	FlashlightPlusHolder setupFlashlightHolder(PlayerPawn p){
		FlashlightPlusHolder holder = FlashlightPlusHolder(p.GiveInventoryType("FlashlightPlusHolder"));
		holder.Init();
		return holder;
	}
	
	override void WorldLoaded(WorldEvent e){
		if(e.IsReopen){
			FlashlightPlusLight hl=null;
			for(let it=ThinkerIterator.Create("FlashlightPlusLight");hl=FlashlightPlusLight(it.next());){
				hl.destroy();//prevents duplicate flashlights for hub-world maps
			}
		}
		for(int i=0;i<MAXPLAYERS;i++){
			if(playeringame[i]){
				PlayerPawn p=players[i].mo;
				FlashlightPlusHolder holder=FlashlightPlusHolder(p.FindInventory("FlashlightPlusHolder"));
				if(holder){
					holder.FixState();
				}
			}
		}
		FlashlightPlusHolder holder=null;
		for(let it=ThinkerIterator.Create("FlashlightPlusHolder");holder=FlashlightPlusHolder(it.next());){
			holder.FixState();
		}
	}
	
	override void PlayerDied(PlayerEvent e){//disable flashlight on player death
		let p=players[e.playerNumber].mo;
		FlashlightPlusHolder holder=FlashlightPlusHolder(p.FindInventory("FlashlightPlusHolder"));
		if(holder){
			holder.Disable();
		}
	}
	
	override void PlayerDisconnected(PlayerEvent e){//reset state on player disconnect
		FlashlightPlusLight hl=null;
		for(let it=ThinkerIterator.Create("FlashlightPlusLight");hl=FlashlightPlusLight(it.next());){
			hl.destroy();
		}
		FlashlightPlusHolder holder=null;
		for(let it=ThinkerIterator.Create("FlashlightPlusHolder");holder=FlashlightPlusHolder(it.next());){
			if(holder.owner&&!playeringame[holder.owner.PlayerNumber()])continue;
			holder.FixState();
		}
	}
	
	override void NetworkProcess(ConsoleEvent e){
		if(e.name=="flashlight_plus_toggle"){
			PlayerPawn p=players[e.player].mo;
			if(p){
				FlashlightPlusHolder holder=FlashlightPlusHolder(p.FindInventory("FlashlightPlusHolder"));
				if(!holder){
					holder=setupFlashlightHolder(p);
				}
				holder.ToggleFlashlight();
			}
		}else if(e.name=="flashlight_plus_uninstall"){
			PlayerPawn p=null;
			for(let it=ThinkerIterator.Create("PlayerPawn");p=PlayerPawn(it.next());){
				FlashlightPlusHolder holder=FlashlightPlusHolder(p.FindInventory("FlashlightPlusHolder"));
				if(holder){
					p.RemoveInventory(holder);
					holder.destroy();
				}
			}
			FlashlightPlusLight hl=null;
			for(let it=ThinkerIterator.Create("FlashlightPlusLight");hl=FlashlightPlusLight(it.next());){
				hl.destroy();
			}
		}
	}
}

class FlashlightPlusLight : Spotlight {
	
	Default {
		+NOINTERACTION;
	}
	
	enum ELocation {
		HELMET = 0,
		RIGHT_SHOULDER = 1,
		LEFT_SHOULDER = 2,
		CAMERA = 3
	};
	
	PlayerPawn toFollow;
	
	Vector3 offset;
	
	FlashlightPlusLight Init(PlayerPawn p,bool second) {
		toFollow=p;
		
		Color c=CVar.GetCVar("cl_flashlight_plus_color",toFollow.player).GetString();
		
		if(second){
			float mult=CVar.GetCVar("cl_flashlight_plus_color_2_mult",toFollow.player).GetFloat();
			args[0]=c.r*mult;
			args[1]=c.g*mult;
			args[2]=c.b*mult;
		}else{
			args[0]=c.r;
			args[1]=c.g;
			args[2]=c.b;
		}
		
		args[3]=CVar.GetCVar(second?"cl_flashlight_plus_intensity_2":"cl_flashlight_plus_intensity",toFollow.player).GetInt();
		
		SpotInnerAngle=CVar.GetCVar(second?"cl_flashlight_plus_inner_2":"cl_flashlight_plus_inner",toFollow.player).GetFloat();
		SpotOuterAngle=CVar.GetCVar(second?"cl_flashlight_plus_outer_2":"cl_flashlight_plus_outer",toFollow.player).GetFloat();
		
		double zBump=toFollow.height / 15.0;
		
		switch(CVar.GetCVar("cl_flashlight_plus_location",toFollow.player).GetInt()) {
		case HELMET:
			offset=(0,0,zBump);
			break;
		case RIGHT_SHOULDER:
			offset=(toFollow.radius,0,-zBump);
			break;
		case LEFT_SHOULDER:
			offset=(toFollow.radius*-1,0,-zBump);
			break;
		case CAMERA:
			offset=(0,0,0);
			break;
		default:
			offset=(0,0,0);
			break;
		}
		
		return self;
	}
	
	override void Tick() {
		A_SetAngle(toFollow.angle,SPF_INTERPOLATE);
		A_SetPitch(toFollow.pitch,SPF_INTERPOLATE);
		SetOrigin(toFollow.pos+(RotateVector((offset.x,offset.y),toFollow.angle-90.0),toFollow.player.viewheight+offset.z),true);
		Super.Tick();
	}
}